module Main where

import Promela.Generator
import SCons.Parser (parseAlgo)
import SCons.Optimizer (optimize)
import System.Environment (getArgs, getProgName)
import System.Exit (exitFailure)
import Control.Monad (when)
import qualified Data.Text as T
import qualified Data.Text.IO as TIO
import Data.Either (isLeft)

parseFile path = parseAlgo <$> readFile path

genPromela :: FilePath -> FilePath -> IO ()
genPromela src dest = do
  parsed <- parseFile src
  case parsed of
    Left err -> putStrLn $ "Invalid input file:\n" ++ show err
    Right scons -> do
      let pml = generateModel (optimize scons)
      TIO.writeFile dest pml
  
main :: IO ()
main = do
  args <- getArgs
  when (length args /= 2) $ do
    name <- getProgName
    putStrLn $ "Usage: " ++ name ++ " <input_scons_file> <output_pml_file>"
    exitFailure
  let [inFile, outFile] = args
  genPromela inFile outFile
