{-# LANGUAGE QuasiQuotes, OverloadedStrings, ScopedTypeVariables #-}
module Promela.ApplyUpdates where

import SCons
import Promela.VarNames
import Promela.Promela
import Promela.FromFile
import qualified Data.Text as T
import Text.Printf (printf)
import Data.List (transpose)

applyName :: Int -> String
applyName i = "apply_updates_rnd" ++ show i

applyAllLdrTS :: RSpec -> T.Text
applyAllLdrTS s = T.concat [
                    [fromFile|src/Promela/fragments/copy_st_ldr.pml|]
                    , [fromFile|src/Promela/fragments/apply_loop.pml|]
                    , T.pack
                      $ pIf [(compactName, pSeq [
                                 "compact_timestamps();"
                                 , T.unpack $ [fromFile|src/Promela/fragments/copy_ts.pml|]
                                              ])
                            ]
                      (Just "skip")
                    , [fromFile|src/Promela/fragments/cleanup_ts.pml|]
                    , [fromFile|src/Promela/fragments/copy_back_ldr.pml|]
                ]
    where
      applyLoopBody = genApplyLoopBody s Timestamped


applyAllTS :: RSpec -> T.Text
applyAllTS s = T.concat [
                    [fromFile|src/Promela/fragments/copy_st.pml|]
                    , [fromFile|src/Promela/fragments/apply_loop.pml|]
                    , T.pack
                      $ pIf [(compactName, pSeq [
                                 "compact_timestamps();"
                                 , T.unpack $ [fromFile|src/Promela/fragments/copy_ts.pml|]
                                              ])
                            ]
                      (Just "skip")
                    , [fromFile|src/Promela/fragments/cleanup_ts.pml|]
                    , [fromFile|src/Promela/fragments/copy_back.pml|]
                ]
    where
      applyLoopBody = genApplyLoopBody s Timestamped


applyAllLdr :: RSpec -> T.Text
applyAllLdr s = T.concat [
                    [fromFile|src/Promela/fragments/copy_st_ldr.pml|]
                    , [fromFile|src/Promela/fragments/apply_loop.pml|]
                    , [fromFile|src/Promela/fragments/copy_back_ldr.pml|]
                ]
    where
      applyLoopBody = genApplyLoopBody s Normal

applyAllLdrRand :: RSpec -> T.Text
applyAllLdrRand s = T.concat [
                    [fromFile|src/Promela/fragments/copy_st_ldr.pml|]
                    , [fromFile|src/Promela/fragments/apply_loop.pml|]
                    , [fromFile|src/Promela/fragments/copy_back_ldr.pml|]
                ]
    where
      applyLoopBody = genApplyLoopBody s Randomized

applyAllRand :: RSpec -> T.Text
applyAllRand s = T.concat [
                    [fromFile|src/Promela/fragments/copy_st.pml|]
                    , [fromFile|src/Promela/fragments/apply_loop.pml|]
                    , [fromFile|src/Promela/fragments/copy_back.pml|]
                ]
    where
      applyLoopBody = genApplyLoopBody s Randomized
 

applyAll :: RSpec -> T.Text
applyAll s = T.concat [
                    [fromFile|src/Promela/fragments/copy_st.pml|]
                    , [fromFile|src/Promela/fragments/apply_loop.pml|]
                    , [fromFile|src/Promela/fragments/copy_back.pml|]
                ]
    where
      applyLoopBody = genApplyLoopBody s Normal
 
genApplyLoopBody :: RSpec -> AlgoType -> T.Text
genApplyLoopBody s aType = T.unlines $ map fieldUpd (zip [1..] iss)
  where
    hasTS = aType == Timestamped
    isRand = aType == Randomized
    iss = instrs s
    nFields = length iss
    updVar :: Int -> String -> String
    updVar i ch
      | nFields == 1  = printf "_topu1(%s)" ch
      | nFields == 2  = printf "_topu%d2(%s)" i ch
    botCheck :: Int -> String -> String
    botCheck i ch
      | nFields == 1  = printf "_bot_pu1(%s)" ch
      | nFields == 2  = printf "_bot_pu%d2(%s)" i ch
    fieldUpd (i, Upd _ _ fName)
      | isEph fName = [fromFile|src/Promela/fragments/upd_ephemeral.pml|]
      | hasTS && fName == inpName = [fromFile|src/Promela/fragments/upd_tsed_inp.pml|]
      | isRand && fName == inpName = [fromFile|src/Promela/fragments/upd_persistent_random.pml|]
      | otherwise = [fromFile|src/Promela/fragments/upd_persistent.pml|]
          where
            updField :: String -> String -> String -> String
            updField iVar stVar chVar =
              let
                (ch :: String) = printf "%s[%s]" chVar iVar
              in
                printf "%s(%s[%s], %s)"
                (updaterName fName) stVar iVar (updVar i ch)
            botChoice chVar iVar =
              let
                (ch :: String) = printf "%s[%s]" chVar iVar
              in
              botCheck i ch
    
            updRField :: String -> String -> String -> String
            updRField iVar stVar chVar =
              let
                (ch :: String) = printf "%s[%s]" chVar iVar
              in
                printf "%s(%s[%s], %s)"
                (updaterName fName) stVar iVar ch


applyLdr :: RSpec -> T.Text
applyLdr s = [fromFile|src/Promela/fragments/apply_ldr.pml|]
  where
    chainStepUpdates st ch =
      foldr (\f st -> updaterName f ++ (printf "(%s, %s[0])" st ch)) st fields
    fields = map updTarget (instrs s)

applyRound :: Algo -> (Int, RSpec) -> T.Text
applyRound a (rnd, s) = [fromFile|src/Promela/fragments/apply_rnd_skel.pml|]
  where
    applyBody
      | commPattern s == AllToLdr   = applyLdr s
      | tsNeeded && hasLdr a        = applyAllLdrTS s
      | tsNeeded                    = applyAllTS s
      | randNeeded && hasLdr a      = applyAllLdrRand s
      | randNeeded                  = applyAllRand s
      | hasLdr a                    = applyAllLdr s
      | otherwise                   = applyAll s
    tsNeeded = isTSed a && (inpName `elem` (map updTarget (instrs s)))
    randNeeded = isRand a && (inpName `elem` (map updTarget (instrs s)))

applyUpdates :: Algo -> T.Text
applyUpdates a = T.concat [compaction, individual, dispatch]
  where
    rnds = a_rounds a
    individual = T.unlines $ map (applyRound a) (zip [0..] rnds)
    compaction
      | isTSed a    = [fromFile|src/Promela/fragments/compact_ts.pml|]
      | otherwise   = ""
    dispatch = [fromFile|src/Promela/fragments/apply_dispatch.pml|]
      where
        cases = T.unlines $ map cs [0..length rnds - 1]
        cs i = T.pack $ printf " :: %s == %d -> %s();" rndVar i (applyName i)
 
