{-# LANGUAGE QuasiQuotes, OverloadedStrings #-}
module Promela.CalculateUpdates
  (generateUpdates)
where  
 
import SCons
import SCons.Utils
import Promela.VarNames
import Promela.FromFile
import Promela.Promela
import qualified Data.Text as T
import Text.Printf (printf)
import Data.Ratio (denominator, numerator)
import Data.List (nub, transpose)

overTh :: Rational -> String -> String
overTh th cnt = printf "(%d * (%s) > %d * N)" (denominator th) cnt (numerator th)
 
reachTh :: Rational -> String -> String
reachTh th cnt = printf "(%d * (%s) >= %d * N)" (denominator th) cnt (numerator th)

-- TODO: where do we handle PhiUnif? For maxTS it doesn't really matter,
-- since it's deterministic...
genMaxTsUpdates :: RSpec -> RFormula -> T.Text
genMaxTsUpdates s phi = [fromFile|src/Promela/fragments/maxts_calc.pml|] 
  where
    is = instrs s
    phiThStr :: String
    phiThStr
      | any isPhiT phi   = "true"
      | otherwise        = "false"
    nFields = length (instrs s)
    possUpds :: String -> String
    possUpds s
      | nFields == 1  = printf "_pupds1(%s)" s
      | nFields == 2  = printf "_pupds2(%s, %s)" s s
    maxTh = maximum $ map th is
    minTh = minimum $ map th is
    overMaxTh = overTh maxTh
    overMinTh = overTh minTh
    mixedBot :: Int -> T.Text
    mixedBot val
      | nFields == 1    = ""
      | any isPhiT phi  = ""
      | minTh == maxTh  = ""
      | otherwise       = [fromFile|src/Promela/fragments/maxts_mixed.pml|]
          where
            possMixUpd :: String
            possMixUpd
              | th (head is) > th (last is) = printf "_pupds2(BOT, %d)" val
              | otherwise                   = printf "_pupds2(%d, BOT)" val

genCntVals :: RSpec -> RFormula -> T.Text
genCntVals s phi = cntUpdates
  where
    sField = sendField s
    cntUpdates :: T.Text
    cntUpdates = [fromFile|src/Promela/fragments/cnt_vals.pml|] 

cond :: Rational -> Predicate -> Int -> String
cond th Any v  = (nVar v ++ " > 0") `pAnd` overTh th (nVar 0 ++ " + " ++ nVar 1)
cond th Smin 0  = (nVar 0 ++ " > 0") `pAnd` overTh th (nVar 0 ++ " + " ++ nVar 1)
cond th Smin 1  = overTh th (nVar 1)
cond th Smor 0  = reachTh (th/2) (nVar 0)
cond th Smor 1  = overTh (th/2) (nVar 1)
cond th AllEq 0  = overTh th (nVar 0)
cond th AllEq 1  = overTh th (nVar 1)

botRcvCond :: Rational -> RFormula -> String
botRcvCond t phi
  | any isPhiT phi && t == 0 = overTh thPhi nBots
  | any isPhiT phi = overTh 0 nBots
  | otherwise       = "true"
  where
    nBots :: String
    nBots = printf "N - (%s + %s)" (nVar 0) (nVar 1)
    thPhi = head [th | PhiT th <- phi]

grdBotCond :: Predicate -> String
grdBotCond AllEq = ((nVar 0 ++ " > 0") `pAnd` (nVar 1 ++ " > 0"))
grdBotCond _ = "false"

botCond :: Rational -> Predicate -> RFormula -> String
botCond t p phi = grdBotCond p `pOr` botRcvCond t phi

setUpds1 :: RSpec -> RFormula -> T.Text
setUpds1 s phi = [fromFile|src/Promela/fragments/set_upds1.pml|]
  where
    is = instrs s
    t = th $ head is
    p = predicate $ head is
    genCond1 v = cond t p v
    possUpds :: String -> String
    possUpds = printf "_pupds1(%s)"
    genBotCond = botCond t p phi

setUpds2 :: RSpec -> RFormula -> T.Text
setUpds2 s phi = T.unlines updList
  where
    updList = [ [fromFile|src/Promela/fragments/set_upds2.pml|] | v1 <- [V0, V1, VBot], v2 <- [V0, V1, VBot]]
    possUpds2 :: BinVal -> BinVal -> String
    possUpds2 v1 v2 = printf "_pupds2(%s, %s)" (show v1) (show v2)
    is = instrs s
    preds = map predicate is
    mixedPossible = Any `elem` preds && AllEq `notElem` preds
    ui1 = head is
    ui2 = last is
    g1Larger = (th ui1, predicate ui1) >= (th ui2, predicate ui2)
    eqGuards = (th ui1, predicate ui1) == (th ui2, predicate ui2)
    maxTh = maximum $ map th is
    minTh = minimum $ map th is
    maxP = maximum preds
    minP = minimum preds
    cond2 V0 V0 = cond maxTh maxP 0
    cond2 V1 V1 = cond maxTh maxP 1
    cond2 V0 V1 
      | not mixedPossible       = "false"
      | maxP == Smin
        && g1Larger             = cond maxTh maxP 0 `pAnd` cond minTh minP 1
      | maxP == Smin            = "false"
      | otherwise               = cond maxTh maxP 0 `pAnd` cond minTh minP 1
    cond2 V1 V0
      | not mixedPossible       = "false"
      | maxP == Smin
        && not g1Larger         = cond maxTh maxP 0 `pAnd` cond minTh minP 1
      | maxP == Smin            = "false"
      | otherwise               = cond maxTh maxP 0 `pAnd` cond minTh minP 1
    cond2 VBot VBot = botCond minTh minP phi
    cond2 v VBot
      | not bot2Possible        = "false"
      | minTh == maxTh          = cond minTh minP iv `pAnd` grdBotCond maxP
      | otherwise               = cond minTh minP iv `pAnd` botCond maxTh maxP phi
      where
        bot2Possible = predicate ui1 /= predicate ui2 && not g1Larger
        iv = case v of
          V0 -> 0
          V1 -> 1
    cond2 VBot v
      | not bot1Possible        = "false"
      | minTh == maxTh          = cond minTh minP iv `pAnd` grdBotCond maxP
      | otherwise               = cond minTh minP iv `pAnd` botCond maxTh maxP phi
      where
        bot1Possible = predicate ui1 /= predicate ui2 && g1Larger
        iv = case v of
          V0 -> 0
          V1 -> 1

genNonTsUpdate :: RSpec -> RFormula -> T.Text
genNonTsUpdate s phi = T.concat [cnt, setUpds, resetCtrs]
  where
    cnt = genCntVals s phi
    is = instrs s
    setUpds = case length is of
                1 -> setUpds1 s phi
                2 -> setUpds2 s phi
    resetCtrs = [fromFile|src/Promela/fragments/reset_ctrs.pml|]

genUpdatesFromLdr :: RSpec -> RFormula -> T.Text
genUpdatesFromLdr s phi = [fromFile|src/Promela/fragments/ldr_update.pml|]
  where
    nFields = length $ instrs s
    sField = sendField s
    possUpds :: String -> String
    possUpds v
      | nFields == 1    = printf "_pupds1(%s)" v
      | nFields == 2    = printf "_pupds2(%s, %s)" v v
    botPossible :: String 
    botPossible
      | PhiLS `notElem` phi = "true"
      | otherwise           = "false"
    ldrVal :: String
    ldrVal = printf "%s(%s)" (decoderName sField) ldrStVarName

genRoundUpdates :: RSpec -> RFormula -> T.Text
genRoundUpdates s@(RSpec sField cp instrs) phi
  | SmaxTS `elem` (specPreds s)     = genMaxTsUpdates s phi
  | commPattern s == LdrToAll       = genUpdatesFromLdr s phi
  | otherwise                       = genNonTsUpdate s phi

genUpdateWithPreds :: Int -> RSpec -> RFormula -> [RFormula] -> T.Text
genUpdateWithPreds rnd s phiS phiTs = [fromFile|src/Promela/fragments/calc_rnd_skel.pml|]
  where
    cases = pIf (map cse (zip [1..] phiTs)) (Just elseCase)
    cse (i, phiT) = (phiName i, T.unpack $ genRoundUpdates s (phiS ++ phiT))
    elseCase = T.unpack $ genRoundUpdates s phiS

generateUpdates :: Algo -> T.Text
generateUpdates a = T.append individual dispatch 
  where
    rnds = a_rounds a
    safety = a_safety a
    term = a_term a
    individual = T.unlines
                 $ map (\(i, (r, s, t)) -> genUpdateWithPreds i r s t)
                 $ zip [0..] $ splitIntoRounds a
    dispatch = [fromFile|src/Promela/fragments/calc_dispatch.pml|]
      where
        cases = T.unlines $ map cs [0..length rnds - 1]
        cs i = T.pack $ printf " :: %s == %d -> %s();" rndVar i (calculateName i)

