module Promela.ChooseUpdates where

import SCons
import SCons.Utils
import Promela.VarNames
import Promela.Promela
import qualified Data.Text as T
import Data.List (transpose)

infixr 3 ==> 

(==>) :: Bool -> Bool -> Bool
(==>) p q = not p || q

choiceVal1 :: Int -> BinVal
choiceVal1 0 = V0
choiceVal1 1 = V1
choiceVal1 2 = VBot

choiceVal2 :: Int -> (BinVal, BinVal)
choiceVal2 i = [(v1, v2) | v1 <- [V0, V1, VBot], v2 <- [V0, V1, VBot]] !! i

chooseProcRndUpdate :: RSpec -> RFormula -> Int -> T.Text
chooseProcRndUpdate s phi pId
  | PhiU `elem` phi && pId == 0     = chooseAndRestrict
  | otherwise                       = choose
  where
    is = instrs s
    nFields :: Int
    nFields = length is
    nrChoices :: Int
    nrChoices = 3 ^ nFields
    maxChoice = nrChoices - 1
    chooseAndRestrict = T.pack $ pIf (map choiceAndRestr [0..maxChoice]) Nothing
    choose = T.pack $ pIf (map choice $ map show [0..maxChoice]) Nothing
    choice :: String -> (String, String)
    choice i = (pCall "_pupds1" [i], pSet (pArr updChoiceName (show pId)) i)
    choiceAndRestr :: Int -> (String, String)
    choiceAndRestr c = (pCall "_pupds1" [show c],
                        pSeq (
                           (pSet (pArr updChoiceName (show pId)) (show c))
                           : map
                             (\i -> pSet (pCall "_pupds1" [show i]) (updatePossible i))
                             [0..maxChoice]
                           -- TODO: finish this!!!
                           ))
      where
        updatePossible i
          | nFields == 1 = if i == c ||
                              ((i /= 2 && c /= 2) && predicate (head is) == Any)
                           then (pCall "_pupds1" [show i])
                           else "false"
          | nFields == 2 = let (v1, v2) = choiceVal2 c; (v1', v2') = choiceVal2 i
                           in
                             if (v1 /= v1' ==> (predicate (is !! 0) == Any &&
                                                v1 /= VBot && v1' /= VBot))
                                && (v2 /= v2' ==> (predicate (is !! 1) == Any &&
                                                v2 /= VBot && v2' /= VBot))
                             then (pCall "_pupds1" [show i])
                             else "false"


chooseAllRndUpdates :: Int -> RSpec -> RFormula -> T.Text
chooseAllRndUpdates n s phi = T.unlines $
                                  map (chooseProcRndUpdate s phi) [0..n-1]

    
chooseUpdates :: Algo -> T.Text
chooseUpdates a = T.pack
                  $ pDefInline "choose_updates"
                  $ pIf (map chooseRnd rndsWithPreds) Nothing
  where
    rnds = a_rounds a
    safety = a_safety a
    term = a_term a
    rndsWithPreds = zip [0..] $ splitIntoRounds a
    n = cutoff a
    
    chooseRnd (i, (s, phiS, phiTs))
      | commPattern s == AllToLdr   = (pEq rndVar (show i),
                                       T.unpack $ chooseProcRndUpdate s [] 0)
      | otherwise                   = (pEq rndVar (show i),
                                       pIf
                                        (map
                                          (\(i, phiT) ->
                                             (phiName i,
                                              T.unpack $
                                               chooseAllRndUpdates n s (phiS ++ phiT)))
                                          unifPhiTs)
                                        (Just $ T.unpack $
                                           chooseAllRndUpdates n s phiS)
                                       )
      
      where
        unifPhiTs = filter (elem PhiU . snd) $ zip [1..] phiTs
    
