{-# LANGUAGE QuasiQuotes #-}
module Promela.Driver where

import SCons
import Promela.Promela
import Promela.FromFile
import Promela.VarNames
import qualified Data.Text as T

chooseInputs :: Algo -> String
chooseInputs a = pBlock "atomic" (pSeq choices)
  where
    n = cutoff a
    choices = map choice [0..n-1]
    choice i = pIf
               [("true", "skip"),
                 ("true", pSeq [
                     "st_ctr[INITIAL]--;"
                     , "st_ctr[upd_inp(INITIAL, 1)]++"
                     ])
               ]
               Nothing

generateDriver :: Algo -> T.Text
generateDriver a = T.pack $ pDefProctype "model" body
  where
    nrSteps = length (a_rounds a)
    body = pSeq [
      pSet (pArr ctrVarName "INITIAL") "N"
      , chooseInputs a
      , mainLoop
      ]
    drawRand
      | isRand a = "draw_randomness();"
      | otherwise = ""
    cleanRand
      | isRand a = "cleanup_randomness();"
      | otherwise = ""
    mainLoop = pDo [("true", pSeq [
                        toggleStart
                        , chooseLdrAndPreds
                        , pBlock "d_step" "calculate_updates();"
                        , pBlock "atomic" ("choose_updates();" ++ drawRand)
                        , pBlock "d_step" (pSeq ["apply_updates();"
                                                , "update_decs();"
                                                , "cleanup();"
                                                , cleanRand
                                                , pSet rndVar
                                                  ("(" ++ rndVar ++ " + 1) % "
                                                    ++ show nrSteps)
                              ])
                        ])
                   ]
               Nothing

    chooseLdrAndPreds = pBlock "atomic"
                        $ pIf [(pEq rndVar "0", pSeq
                                 [chooseLeader, choosePreds, chooseLuckyPred])]
                        (Just "skip")
    chooseLeader
      | hasLdr a    = T.unpack $ [fromFile|src/Promela/fragments/choose_ldr.pml|]
      | otherwise   = ""
    term = a_term a
    choosePreds = pSeq $ map choosePhiT [1..length term]
    choosePhiT i = pIf [("true", pSet (phiName i) "false"),
                        ("true", pSet (phiName i) "true")]
                   Nothing

    chooseLuckyPred
      | isRand a  = pIf [("true", pSet "phi_lucky" "false"),
                         ("true", pSet "phi_lucky" "true")]
                    Nothing
      | otherwise = ""

    toggleStart
        | length term > 1   = pIf
                              [(pEq rndVar "0",
                                pSeq [pSet "start" "true", pSet "start" "false"])]
                              (Just ("skip"))
        | otherwise         = ""
                                
