module Promela.Promela where

import Text.Printf (printf)
import Data.List (intercalate)


pNot :: String -> String
pNot = printf "!(%s)"

pAnd :: String -> String -> String
pAnd = printf "(%s) && (%s)"

pOr :: String -> String -> String
pOr = printf "(%s) || (%s)"

pImp :: String -> String -> String
pImp p q = pNot p `pOr` q

pEq :: String -> String -> String
pEq = printf "%s == %s"

pNeq :: String -> String -> String
pNeq = printf "%s != %s"

pChoice :: String -> String -> [(String, String)] -> Maybe String -> String
pChoice stDelim endDelim [] Nothing = ""
pChoice stDelim endDelim [] (Just st) = st
pChoice stDelim endDelim cases fallback =
  case fallback of
    Nothing -> printf "%s \n\
                      \%s\n\
                      \%s;\n" stDelim (unlines $ map (indent . gen) cases) endDelim
    Just st -> printf "%s\n\
                      \%s\n\
                      \:: else ->\n\
                      \%s\n\
                      \%s;\n"
               stDelim
               (unlines $ map (indent . gen) cases)
               (indent $ indentBy 3 st)
               endDelim
  where
    gen (cond, st) = printf ":: %s ->\n\
                            \%s" cond (indent st)

pIf :: [(String, String)] -> Maybe String -> String
pIf = pChoice "if" "fi"

pDo :: [(String, String)] -> Maybe String -> String
pDo = pChoice "do" "od"

indentBy :: Int -> String -> String
indentBy i cmds = unlines $ map ((replicate i ' ')++) $ lines cmds

indent :: String -> String
indent cmds = unlines $ map ("  "++) $ lines cmds

pDefInline :: String -> String -> String
pDefInline name cmds = printf "inline %s(){\n\
                              \%s\n\
                              \}\n" name (indent cmds)

pDefProctype :: String -> String -> String
pDefProctype name body = printf "active proctype %s(){\n\
                                \%s\n\
                                \}\n" name (indent body)

pCall :: String -> [String] -> String
pCall f args = printf "%s(%s)" f (intercalate ", " args)

pArr :: String -> String -> String
pArr arrVar index = printf "%s[%s]" arrVar index

pSet :: String -> String -> String
pSet var val = printf "%s = %s;" var val

pSeq :: [String] -> String
pSeq = unlines

pDeclare :: String -> String -> String
pDeclare typ name = printf "%s %s;" typ name

pMacro :: String -> [String] -> String -> String
pMacro name args body = case args of
                          [] ->
                            printf "#define %s %s" name body
                          _ ->
                            printf "#define %s(%s) %s" name (intercalate ", " args) body
  
pBlock :: String -> String -> String
pBlock header body = printf "%s {\n\
                            \%s\n\
                            \}\n" header (indent body)

pSGlobally :: String -> String
pSGlobally p = printf "([] (%s))" p

pSEventually :: String -> String
pSEventually p = printf "<> (%s)" p

pSNext :: String -> String
pSNext p = printf "X (%s)" p

pSImp :: String -> String -> String
pSImp p q = printf "(%s) -> (%s)" p q
  
