module SCons where

import Data.Ratio (denominator)

data CommPattern = AllToAll | LdrToAll | AllToLdr
  deriving (Show, Eq)

type FName = String

data Predicate = Any | Smor | Smin | SmaxTS | AllEq
  deriving (Show, Eq)

instance Ord Predicate where
  Any <= _   = True
  _ <= AllEq = True
  x <= y
    | x == y    = True
    | otherwise = False

type Threshold = Rational

data Guard = G (Threshold, Predicate)
  deriving (Show, Eq)


data UpdInstr = Upd {
  th :: Threshold
  , predicate :: Predicate
  , updTarget :: FName
  }
  deriving (Show, Eq)

data RSpec = RSpec {
  sendField :: FName
  , commPattern :: CommPattern
  , instrs :: [UpdInstr]
  }
  deriving (Show, Eq)

af :: Int -> Threshold -> Int
af n t = floor (fromIntegral n * t) + 1

specPreds :: RSpec -> [Predicate]
specPreds s = map predicate (instrs s)

specUpdFields :: RSpec -> [FName]
specUpdFields s = map updTarget (instrs s)

data CommFormula = PhiU | PhiT Rational | PhiLS | PhiLR Rational
  deriving (Eq, Show)

type RFormula = [CommFormula]

data AlgoType = Normal | Timestamped | Randomized
  deriving (Eq, Show)

data Algo = Algo {
  a_type :: AlgoType
  , a_rounds :: [RSpec]
  , a_safety :: [RFormula]
  , a_term :: [[RFormula]]
  }
  deriving (Show)

isTSed :: Algo -> Bool
isTSed (Algo Timestamped _ _ _) = True
isTSed _ = False

isRand :: Algo -> Bool
isRand (Algo Randomized _ _ _) = True
isRand _ = False

hasLdr :: Algo -> Bool
hasLdr (Algo _ ss _ _) = any (\s -> commPattern s `elem` [LdrToAll, AllToLdr]) ss

isPhiT :: CommFormula -> Bool
isPhiT (PhiT _) = True
isPhiT (PhiLR _) = True
isPhiT _ = False

cutoff :: Algo -> Int
cutoff a = 2 * d + 1
  where
    d = fromInteger $ maximum $ concatMap (map (denominator . th) . instrs) (a_rounds a)

type FVal = Int
type AbsVal = Int

inpName :: FName
inpName = "inp"

decName :: FName
decName = "dec"

isEph :: FName -> Bool
isEph f = notElem f [inpName, decName]

data BinVal = V0 | V1 | VBot
  deriving Eq

instance Show BinVal where
  show V0 = "0"
  show V1 = "1"
  show VBot = "BOT"
  



