module SCons.Optimizer where

import SCons
import SCons.Utils
import Data.List (intersect, nub, (\\), sortOn)

ephName = "eph"

fieldsLeadingTo :: FName -> [RSpec] -> [FName]
fieldsLeadingTo _ [] = []
fieldsLeadingTo fn (r:rs) 
  | fn `elem` ufs                   = [sf]
  | not (null (intersect ufs next)) = sf : next 
  | otherwise                       = next
  where
    sf = sendField r
    ufs = specUpdFields r
    next = fieldsLeadingTo fn rs

inpPath :: [RSpec] -> [FName]
inpPath = fieldsLeadingTo inpName 

decPath :: [RSpec] -> [FName]
decPath s = fieldsLeadingTo decName s \\ fieldsLeadingTo inpName s

swapElems :: [a] -> Int -> Int -> [a]
swapElems (x:xs) 0 0 = (x:xs)
swapElems (x:xs) 0 n = let (f, (y:s)) = splitAt (n-1) xs in y : f ++ (x : s)
swapElems (x:xs) n 0 = let (f, (y:s)) = splitAt (n-1) xs in y : f ++ (x : s)
swapElems (x:xs) n m = x : swapElems xs (n-1) (m-1)

-- collapseEph :: Algo -> Algo
-- collapseEph a = a { a_rounds = collapsedEph False (a_rounds a) }
--   where
--     collapsedEph False (r:rs)
--       | nInstrs == 1    = undefined
--       | otherwise       = collapsedEph True rs

optimizePhase :: Algo -> Algo
optimizePhase a = fromRounds (a_type a) . collapseEph . reorder $ splitIntoRounds a
  where
    sp = splitIntoRounds a
    newSp = reorder sp
    reorder = sortOn
              (\(r, _, _) -> if sendField r `elem` inpPath (a_rounds a) then 0 else 1)
    collapseEph [] = []
    collapseEph (r@(rnd@(RSpec sf cp is), s, t):rps) = newR : collapseEph newRs
      where
        subs = take 1 [(f, sf) | f <- specUpdFields rnd, isEph f, isEph sf]
        (fr, to) = head subs
        newR
          | length subs == 1 = (renameSpecField fr to rnd, s, t)
          | otherwise        = r
        newRs
          | length subs == 1 = renameFieldGlobal fr to rps
          | otherwise = rps
        
optimize :: Algo -> Algo
optimize = foldr1 concatAlgos . map optimizePhase . splitIntoPhases
