module SCons.Utils where

import Data.List (nub, zip3, transpose)
import SCons

allDiff :: Eq a => [a] -> Bool
allDiff l = length l == length (nub l)

splitWhen :: (a -> Bool) -> [a] -> [[a]]
splitWhen p [] = [[]]
splitWhen p (x:xs)
  | p x = [] : (x : head rest) : tail rest
  | otherwise = (x : head rest) : tail rest
  where
    rest = splitWhen p xs

type RWithPhi = (RSpec, RFormula, [RFormula])

splitIntoRounds :: Algo -> [RWithPhi]
splitIntoRounds a = zip3 (a_rounds a) (a_safety a) tterms
  where
    tterms
      | null (a_term a) = replicate (length (a_rounds a)) []
      | otherwise = transpose (a_term a)

fromRounds :: AlgoType -> [RWithPhi] -> Algo
fromRounds typ rdesc = Algo typ rs s ts
  where
    (rs, s, ts') = unzip3 rdesc
    ts = transpose ts'
    
splitIntoPhases :: Algo -> [Algo]
splitIntoPhases a = map (fromRounds (a_type a)) split
  where
    split = splitWhen (\(r, _, _) -> sendField r == inpName) (splitIntoRounds a)

concatAlgos :: Algo -> Algo -> Algo
concatAlgos a1 a2 = if (a_type a1 /= a_type a2)
                   then error "concatAlgos: Different algorithm types"
                   else fromRounds (a_type a1)
                        (splitIntoRounds a1 ++ splitIntoRounds a2)

renameSpecField :: FName -> FName -> RSpec -> RSpec
renameSpecField from to (RSpec sf cp is) = RSpec (sub sf) cp (mapTargets sub is)
  where
    mapTargets f = map (\x -> x { updTarget = f (updTarget x) })
    sub s
      | s == from   = to
      | otherwise   = s

renameFieldGlobal :: FName -> FName -> [RWithPhi] -> [RWithPhi]
renameFieldGlobal from to l = [(renameSpecField from to r, s, t) | (r, s, t) <- l]

splitRndsIntoPhases a = if null (head parts)
                    then Right (tail parts)
                    else Left "Invalid start of phase; must send inp"
  where
    parts = splitWhen (\x -> sendField x == inpName) rnds
    rnds = a_rounds a

count :: Eq a => a -> [a] -> Int
count e l = length [1 | y <- l, y == e]

