This is a proof-of-concept tool for generating Promela models of SCons
algorithms.

The input syntax of the tool does not cover the full SCons
language. In particular, the phase sequence and communication
predicate specifications are subsets of the ones in the paper, and the
support for randomized algorithms is limited. Also, the input syntax
is ASCII and thus slightly different from the BNF in the paper.

The best way to understand the syntax is probably by looking at the
scons-examples directory; it contains the SCons code for all our case
studies.


Phase Sequences
---------------
For the phase sequence, we assume that the algorithm is given by a
list of rounds that keeps repeating forever. That is, we only allow
sequences of type v^\omega instead of uv^\omega.

Communication predicates
------------------------

For communication predicates, the alphabet is the same as in the
paper, but we do not allow arbitrary omega-regular expressions. Instead,
we assume that the predicate is defined by two components:

  * a list of _safety_ formulas, where the length of the list matches
    the number of the rounds of the algorithm. The i-th safety formula
    in the list must always hold in the i-th round. The list is prefixed
    by the keyword SAFETY. See scons-examples/cuv.scons for an example.

  * a list of lists of _termination_ formulas. Again, the length of each
    (sub)list must match the number of rounds of the algorithm. The semantics
    is then that, given lists T1, T2... TN of lists of termination formulas,
    T1 must hold in a phase, then T2 must hold in some subsequent phase,
    and so on. See scons-examples/otr.scons for the example described in
    the paper.
  
Verifying the models
--------------------

Read the INSTALL file for the installation instructions. Here's a possible
example, once you have everything installed:

# Generate the Promela model
./.stack-work/install/x86_64-linux/lts-7.0/8.0.1/bin/scons-gen \
        scons-examples/paxos.scons paxos.pml
# Run spin to generate a C file
spin -a paxos.pml
# A bunch of flags; limit to 29GB of RAM, use hash compaction, no
# partial order reduction
gcc -DMEMLIM=29000 -O2 -DHC4 -DXUSAFE -DCOLLAPSE -DNOREDUCE -w -o pan pan.c
# Depth limit of 2500000
./pan -m2500000 -n -N agreement
./pan -m2500000 -a -N termination
